#include "common.h"
#include "uarts.h"
#include "abcio.h"
#include <plib.h>

s_card* pCard[64];					//Array of pointers to each card struct
//volatile uint32 selectedCard;		//Just a copy of the variable used in isr
volatile uint32 ioBuf[IOBUFSIZE];	//(MSB) 0, data, strobe, cardNum (LSB)
volatile uint32 ioHead = 0;
volatile uint32 ioTail = 0;			//current index in ioBuf[]
s_inp7 cardsInp7[3];
volatile uint8 ioStuff[3] = {0,0,0}; //inp7flag, cardNum, inp1flag
uint16 ioprint = 0;

//void printH8(uint8 c);

//---------------- Initiate all implemented cards here --------
void abcioInit(void)
{
	//The cardsInp7 pointers must be valid!
	cardsInp7[0].pInp7 = (volatile uint8*) &iostuff0_inp7flag;
	cardsInp7[0].inp7val = 1;	//isr will now set inp7flag=1 when inp7 comes
	cardsInp7[1].pInp7 = (volatile uint8*) &iostuff0_inp7flag;
	cardsInp7[1].inp7val = 1;	//must be valid
	cardsInp7[2].pInp7 = (volatile uint8*) &iostuff0_inp7flag;
	cardsInp7[2].inp7val = 1;	//must be valid

	//initialize each card that is implemented - call your init function here
	myCardInit();
	pCard[7] = &myCard;	//select card number
	//while((&myCard.out5stat - myCard) != 26); //alignment test

	pCard[36] = &disk36;	//select card number
	if(SD_DETECT_PIN == 1) disk36_close();	//else wait for Open
	//uninitilized pCard pointers are 0 by default
	
	//Max 3 cards can have inp1 set in isr when inp7 comes
	cardsInp7[0].pInp7 = disk36.pInp1;
	cardsInp7[0].inp7val = 0;
}

//--------------------------------------------------------------
void abcio_disk_open(void)
{
	disk36_open();
}

//--------------------------------------------------------------
void abcio_disk_close(void)
{
	disk36_close();
}

//---------------------------------------------
void delay_us(uint16 ticks)
{	//one tick is 0.8 s
/*	if(ticks==0) return;
	T4CONCLR = 0x8000;
	TMR4=0;
	PR4 = ticks; // + ticks>>2
	IFS0CLR = (1<<16);	//T4IF
	T4CONSET = 0x8000;
	while(! IFS0bits.T4IF);
*/
	uint32 tt=0;
	uint32 t = TMR4;
	while(ticks > tt) tt = TMR4 - t;
}

//--------------------------------------------------------------
void abcioCheck(void)
{	//call this regularly to handle io buffer
	static uint16 repeatedInp1 = 0;
	uint32 io;
	uint8 cardNum, strobe;
	s_card *pSelectedCard = 0;
	uint16 i=0;

	if(ioTail != ioHead)
	{
		do
		{
			io = ioBuf[ioTail];
			ioTail = (ioTail + 1) & (IOBUFSIZE-1);
			strobe = io>>8;
			cardNum = (io >> 16);
			if(ioprint == 1)
			{
				putcPC(CR);	//putcPC(LF);
				putcPC('i'); putcPC('o');
				sendHexPC(cardNum);
				putcPC(strobe + '0');
				sendHexPC(io);
			}
			pSelectedCard = pCard[cardNum & 63];
			if(strobe == STROBE_INP7)
			{	//INP7 - reset to all cards
				//***** call any card function that needs it *****
				//pCard[7]->cardFunc(STROBE_INP7, 0);
				pCard[36]->cardFunc(STROBE_INP7, 0);
				iostuff0_inp7flag = 0;
				continue;
			}
			if(pSelectedCard != 0) pSelectedCard->cardFunc(strobe, io);
		 	if (++i > 300) return; 	//Protect from endless loop if we get bombarded
		}while(ioTail != ioHead);
		iostuff2_inp1flag = 0;
		repeatedInp1 = 0;
	}
	else if(iostuff2_inp1flag != 0)
	{
		iostuff2_inp1flag = 0;
		if(++repeatedInp1 > 2)
		{
			if(pSelectedCard != 0) pSelectedCard->cardFunc(STROBE_INP1X, 0);	//repeated INP1
			repeatedInp1 = 0;
		}
	}

}

//end
