//Header file for ABCbus io card implementations
//---------------------------------------------
/*
IO is handled in assembler that uses the card struct below. 
All io is logged in ioBuf[] even if card is not implemented.
The ioBuf[] is polled regularly.
*/

#include "my_typedefs.h"

#define STROBE_OUT0	(0)

#define STROBE_OUT1	(1)
#define STROBE_CS	(1)

#define STROBE_OUT2	(2)
#define STROBE_C1	(2)

#define STROBE_OUT3	(3)
#define STROBE_C2	(3)

#define STROBE_OUT4	(4)
#define STROBE_C3	(4)

#define STROBE_OUT5	(5)
#define STROBE_C4	(5)

#define STROBE_INP7	(7)
#define STROBE_INP0	(8)
#define STROBE_INP1	(9)
#define STROBE_INP1X (10)
//strobe inp1x given by abcio.c for repeated identical inp1x

#define cardFnParams	uint8 strobe, uint8 data

typedef struct {
	volatile uint8* volatile pInp0;	//volatile pointer to next inp0 data
	uint8* pInp0end;				//pointer to inp0 data buffer end
	volatile uint8* pInp1;			//pointer to volatile inp1 status

	uint16 inp0stat;		//status = (status & LSB) | MSB when inp0 comes.
	uint16 out0stat;		// -"- when out# comes	

	uint16 out2stat;		// -"- when C1# comes	
	uint16 out3stat;		// -"- when C2# comes

	uint16 out4stat;		// -"- when C3# comes
	uint16 out5stat;		// -"- when C4# comes

	void (*cardFunc)(cardFnParams);	//pointer to function that handles strobe and data
	volatile uint8 myReg[8];	//can be used for inp0/inp1/out0 or anything
} s_card;

typedef struct {
	volatile uint8* 	pInp7;
	uint8				inp7val;
} s_inp7;

extern s_card* pCard[64];					//Array of pointers to each card struct
#define IOBUFSIZE 512
extern volatile uint32 ioBuf[IOBUFSIZE];	//(MSB) 0, data, strobe, cardNum (LSB)
extern volatile uint32 ioHead;				//current write position in ioBuf[]
extern volatile uint32 ioTail;				//current read position in ioBuf[]
extern s_inp7 cardsInp7[3]; 				//Cards to set status directly in ISR when inp7 comes
extern uint16 ioprint;
extern volatile uint8 ioStuff[3];			//Various stuff
#define iostuff0_inp7flag	ioStuff[0]		//card reset occured, is set to 0xff when inp7 comes
#define iostuff1_cardNum	ioStuff[1]		//currently selected card, is set to 0xff when inp7 comes (although card 0 is used).
#define iostuff2_inp1flag	ioStuff[2]		//is set to 0xff when a repeated inp1 is not written to ioBuf for implemented card.
//currently selected card is just a copy, i.e. we cannot select card by changing the value.

//----------------------- functions called from main -----------------------------

void abcioInit(void);					//call once
void abcioCheck(void);					//call in main loop to handle abcbus io
void abcio_disk_open(void);
void abcio_disk_close(void);

//------------------- Declare implemented cards here -----------------------------

s_card myCard;	//myCard.c
void myCardInit(void);
void myCardFunc(cardFnParams);

s_card disk36;		//disk.c
void disk36_open(void);
void disk36_close(void);
void diskFunc(cardFnParams);

//end
