;
; This is the ABC80 device driver for the SD0 device, by RJ.
;
; Command
;
; 1:    Open file
; 2:	Prepare file
; 3:	Close file
; 4:	Read block
; 5:	wRITE BLOCK
;
Devlst   EQU             0FE0Ah          ; Device list root
Varlst	 EQU		     0FE29h		; var list at address 65065
Buf0     EQU             0F500h          ; First disk buffer
Closef   EQU             0023H          ; Close file subroutine
RANDJPTAB EQU			28666  ; jump table for random access
UFDINITADDR EQU			06543h ;// in ufd-dos this is c3 43 65
Bprint	EQU		001BH		; print
Binput	EQU		0015H		; input
PROUT	EQU		000BH		; print out routine: text in HL, length in BC


SDBUF	EQU		16384+1024	; 256 byte buffer
Q0BUF	EQU		SDBUF+256	; 256 bytes for command
SEROUTBUF	EQU		SDBUF+512   ; 256 bytes for serial buffer
SERINBUF EQU	SDBUF+768
USBINBUF EQU SDBUF+1024
USBOUTBUF EQU SDBUF+1280

ENDOFBUF EQU	USBOUTBUF+256  ; end of buffers
SDFNAM	EQU		SDBUF-12   ; vars grow downwards
SDFNO	EQU		SDBUF-13
SDSTAT	EQU		SDBUF-14
SDCMD	EQU		SDBUF-15
BLK		EQU		SDBUF-17
BLKSZ	EQU		SDBUF-18
SERINP1	EQU		SDBUF-19   ; circular in buffer read pointer
SERINP2	EQU		SDBUF-20   ; circuilar in buffer write pointer
SEROUTP1 	EQU		SDBUF-21   ; serial out byte read pointer
SEROUTP2	equ		SDBUF-22   ; serial out byte read pointer
SERERR		EQU   	SDBUF-23  ; error byte
EXECADDR	EQU		SDBUF-24  ; exec address for load routine
USBINP1		EQU		SDBUF-25
USBINP2		EQU		SDBUF-26
USBOUTP1		EQU		SDBUF-27
USBOUTP2		EQU		SDBUF-28
CMDMAX	EQU		252



CMDOPEN		EQU		1	;// 1:    Open file
CMDPREP		EQU		2  	;//; 2:	Prepare file
CMDCLOS		EQU		3	;//; 3:	Close file
CMDBLIN		EQU		4	;//; 4:	Read block
CMDBLOUT	EQU		5	;//; 5:	wRITE BLOCK
CMDSEEK		EQU		6

CMDCOPEN	EQU		10
CMDCCLOS	EQU		11
CMDCPRIN	EQU		12
CMDCINPU	EQU	    13

CMDINIT		EQU		99	;//; 99: (re)initialize


ERROPEN	EQU		21		; Errors. Can maybe find better values more aligned with Basic???
ERRCLOS	EQU		46
ERRBLIN	EQU		47
ERRBLOUT EQU	48
ERRNOBLK EQU	49
ERRCOPEN EQU    60
ERRCPRIN EQU	61
ERRCINPU EQU	62
ERRCINLEN EQU	63
ERRCCLOS  EQU	64
ERRSEEK	EQU 	65
ERRLOAD	EQU		66

ERRINIT	EQU		99
ERREOF	EQU		34


;
        ORG             16384   			; 16384 first byte of external ROM - initialize routine
start:  JP              Init				; 16384 jump to init routine
util:   JP				PrintStat			; 16387 jump to print status routine
		JP				PrintQ0buf          ; 16390 print contents of Q0 buffer
		JP				BlockinRaw	        ; 16393 read blk
		JP				BlockoutRaw   		; 16396 write blk
		JP				Seek				; 16399 Seek
		JP				testFindQ0			; 16402 test if Q0 can be found with standard procedure
		JP				SerAddChar			; 16405 add character in E to buffer
		JP				SerPrintChars		; 16408 print out all characters currently in buffer
		JP				SerPrintOutBuf		; 16411 print serial out buffer
		JP				RANDSETFNO			; 16414 set fno (dup of jmp routine at 28666)
		JP				LOADFILE			; 16417 load file into memory at address specified by DE
		JP				SETEXEC				; 16420 set exec used by loadfile =0 means no exec
		JP				USBAddChar			; 16423 add character in E to buffer
		JP				USBPrintChars		; 16426 print out all characters currently in buffer
		JP				USBPrintOutBuf		; 16429 print serial out buffer
		JP				SerGetChar			; 16432 get char in L, H>0 if no char or 0 if char
hellomsg:
		DEFB			'%%%% ABCBus Interface %%%%',13,10
;


; Initialization routine.
; Install the entrys above in the device list
;
Init:
		LD				A,(SDSTAT)
		BIT				7,A
		JR				NZ,init  ; wait until not busy
		LD				A,CMDINIT
		LD				(SDCMD),A
		LD				A,128
		LD				(SDSTAT),A
init2:
		LD				A,(SDSTAT)
		BIT				7,A
		JR				NZ,init2
		CP				0		 ; check if error
		JR				Z,initOK
		RST				2
		DEFB			ERRINIT+128
initok:
        LD				HL,(Devlst) ;   Old devlist <- CMDdev <- Defdev <- SD0dev <- Devlst
		LD              (CMDdev),HL
        LD              HL,CMDdev
		LD              (Defdev),HL
        LD              HL,Defdev
        LD              (SD0dev),HL
        LD              HL,SD0dev
        LD              (Devlst),HL ; first link into dev list
		LD				HL,HELLOMSG
		LD				BC,28
		CALL			PROUT
		JP				UFDDOSINIT ; either RET or patched in ufddos init address


;
; Jumptable at device driver entry
;
SD0tab:
        JP              Open
        JP              Prepare
        JP              Close
        JP              Binput          ; Input
        JP              Bprint          ; Print
        JP              Blockin
        JP              Blockout

CMDtab:
        JP              COpen
        JP              CPrepare
        JP              CClose
        JP              CInput          ; Input
        JP              CPrint          ; Print
        JP              CBlockin
        JP              CBlockout
;
; Open a file for reading and initialize
; the file info block.
;
Open:
		LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,Open   	; Wait for unit to become idle (not executing)
		LD		A,CMDOPEN			; command 1 to open
		LD		(SDCMD),A	; load the command buffer with command

Open2:
		LD      A,(IX+2)    ; File number
		; don't mess with file number now, but PIC will interpret file > 7 as 0
		;CP		8			; larger than 7?
		;JP		NC,Open2b   ; No, continue
		;LD		A,0			; Yes, set to 0 and update (IX+2)
		;LD		(IX+2),A	; store back in parameter block
Open2b:
		LD		(SDFNO),A	; write to register
		EX		DE,HL		; Filename addr in HL (DE points to filename at call to Open routine)
		LD		DE,SDFNAM	; target filename address in DE
		LD		C,11		; 11 characters
		LD		B,0
		LDIR				; write out the filename to the SDFNAM buffer
		LD		A,128		; set bit 7 of sdstat to execute command and error bits to zero
		LD		(SDSTAT),A	; Do it!
 Uopen1:
		LD		A,(SDSTAT)  ; Get result status - pic modifies SDSTAT to signal it is ready
		BIT		7,A			; cmd exec bit is bit 7
		JR		NZ,Uopen1	; wait until not executing
		CP      0           ; other bits zero = OK?
        JR      Z,UOpenOK
        RST     2           ; Error
        DEFB    ERROPEN+128     ; ERR 21 (Can't find file)
		LD 		B,(IX+2)	; Find the buffer to use
UOpenOK:
		LD      HL,Buf0
ULoop1:
		INC     H
        DJNZ    ULoop1
        ; Init device info block
        LD      (IX+7),132      ; Line length (not used)
        LD      (IX+8),L        ; Buffer address
        LD      (IX+9),H
        LD      (IX+0AH),L      ; Read/write position
       	LD      (IX+0BH),H
        LD      (IX+0DH),253    ; No of free chars in buffer
        LD      (IX+0EH),0      ; Buffer is clean
        LD      (HL),3          ; Write EOT at start
        AND     A               ; Clear carry
        RET

;
; Open a file for writing (mostly the same as Open).
;
Prepare:
		LD		A,(SDSTAT)
		BIT		7,A			; wait until not busy
		JR		NZ,Prepare 	; Wait for unit to become idle (not executing)
		LD		A,CMDPREP			; prepare command
		LD		(SDCMD),A	; load the command buffer with command
	    JP      Open2       ; after this then same as open

;
; Close a file
;
Close:
        LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,Close 	; wait until not busy
		BIT     7,(IX+0EH)  ; Buffer dirty?
        CALL    NZ,Closef   ; Call BASIC's "close file"
		LD		A,CMDCLOS			; close command
		LD		(SDCMD),A
		LD      A,(IX+2)    ; File number
		LD		(SDFNO),A
		LD		A,128
		LD		(SDSTAT),A	; execute close command
Close1:
		LD		A,(SDSTAT)
		BIT		7,A			; busy bit is bit 0
		JR		NZ,Close1	; wait until not busy
	    CP      0           ; OK?
        JR      Z,CloseOK
        RST     2           ; Error
        DEFB    ERRCLOS+128    ; Error
CloseOK:
		AND     A           ; Clear carry
        RET

;
; Read a block from a file
;
Blockin:
		LD		A,253
		LD		(BLKSZ),A    ; BASIC block size
		LD      A,(IX+2)        ; File number
        LD		(SDFNO),A
		CALL	BlockinRaw
		JP		BlinBASIC

BlockinRaw:
		LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,Blockin ; wait until not busy
		LD		A,CMDBLIN		; read block command
		LD		(SDCMD),A
		LD		A,128
		LD		(SDSTAT),A		; exec command
Blin1:
		LD		A,(SDSTAT)      ; Get result status
		BIT		7,A
		JR		NZ,Blin1		; wait until not busy
		CP      0               ; OK?
		JR      Z,BlinOK
		CP		ERREOF				; Error number 34 = EOF. Just preserve in A and treat as OK until end of routine
		JR		Z,BlinOK
		RST     2               ; Error
		DEFB    ERRBLIN+128         ; ERR
BlinOK:
		RET

BlinBASIC:
		LD      L,(IX+8)        ; Get buffer address
		LD      H,(IX+9)
		EX		DE,HL		; maybe D and E doesnt work with the IX+xx instructions?
		LD		HL,SDBUF
		LD		B,0
		LD		C,253		; 253 bytes to read into block
		LDIR				; read buffer into disk buffer pointed to by HL
		LD		(HL),3		; write ETX at end at each block
		LD      L,(IX+8)    ; Get buffer address in HL (again)
        LD      H,(IX+9)	; (as denoted in "Avancerad programmering p ABC80", page 163-164)
		CP		ERREOF		; again see if Error was = 3 (EOF). Note A should be preserved from above Blin1
		JR		Z,BlinEOF ;
		AND		A		; clear carry
		RET
BlinEOF:
		XOR		A		; Signal EOF according to "Avancerad Programmering p ABC80, page 162": clear A and set carry flag
		SCF
		RET
;
; Write a block to a file
;
BlockoutRaw:
    	LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,BlockoutRaw ; wait until not busy
		LD		A,CMDBLOUT			; write block command
		LD		(SDCMD),A
		LD		A,128
		LD		(SDSTAT),A	; exec command
BloutRaw1:
		LD		A,(SDSTAT)  ; Get result status
		BIT		7,A
		JR		NZ,BloutRaw1	; wait until not busy
        CP      0           ; OK?
        JR      Z,BloutRawOK
	;	CP		34
	;	JR		Z,BloutRawOK ; Ok also with EOF
        RST     2           ; Error
        DEFB    ERRBLOUT+128    ; Error
BloutRawOK:
		RET

Blockout:
		LD		A,253
		LD		(BLKSZ),A   ; basic block size
    	LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,Blockout ; wait until not busy
		LD		A,CMDBLOUT			; write block command
		LD		(SDCMD),A
		LD      A,(IX+2)    ; File number
        LD		(SDFNO),A
		LD      L,(IX+8)    ; Get buffer address
        LD      H,(IX+9)
		LD		DE,SDBUF
		LD		B,0
		LD		C,253		; 253 bytes to write out
		LDIR				; read buffer into disk buffer pointed to by HL
		LD		A,128
		LD		(SDSTAT),A	; exec command
Blout1:
		LD		A,(SDSTAT)  ; Get result status
		BIT		7,A
		JR		NZ,Blout1	; wait until not busy
        CP      0           ; OK?
        JR      Z,BloutOK
        RST     2           ; Error
        DEFB    ERRBLOUT+128    ; Error
BloutOK:

        LD      (IX+0DH),253    ; No of free chars in buffer
        SET     0,(IX+0EH)   	; Buffer is clean
		LD      L,(IX+8)    ; Get buffer address in HL
        LD      H,(IX+9)
        EX      DE,HL           ; ...and transfer to DE (as denoted in "Avancerad programmering p ABC80", page 163-164)
        AND		A				; clear carry
		RET

Seek:
    	LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,Seek ; wait until not busy
		LD		A,CMDSEEK			; write block command
		LD		(SDCMD),A
		LD		A,128
		LD		(SDSTAT),A	; exec command
Seek1:
		LD		A,(SDSTAT)  ; Get result status
		BIT		7,A
		JR		NZ,Seek1	; wait until not busy
        CP      0           ; OK?
        JR      Z,SeekOK
        RST     2           ; Error
        DEFB    ERRSEEK+128    ; Error
SeekOK:
		RET



		org SDBUF-256

		;
; Define an entry in device list.
;


SD0dev:
        DEFW            0
        ASCII           'SD0'
        DEFW            SD0tab
Defdev:
        DEFW            0
        ASCII           '   '
        DEFW            SD0tab
CMDdev:
        DEFW            0
        ASCII           'CMD'
        DEFW            CMDtab

Q0data:
		DEFB			2	; string type, index 0 (high nibble=0)
		DEFB			'Q'
Q0datanext:
		DEFW			0  ; next PTR
		DEFW			253 ; dimensioned length of Q0$
		DEFW			Q0BUF ; pointer to data - points directly into Q0BUF!
		DEFW			253 ; actual length of Q0$


; here comes routines for command device and random file access
		org				ENDOFBUF   ; routines lie after the buffers

CPrepare:
COpen:
		LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,COpen   	; Wait for unit to become idle (not executing)
		LD		A,CMDCOPEN	; command to open CMD channel
		LD		(SDCMD),A	; load the command buffer with command
		LD		A,128		; set bit 7 of sdstat to execute command and error bits to zero
		LD		(SDSTAT),A	; Do it!
Copen1:
		LD		A,(SDSTAT)  ; Get result status - pic modifies SDSTAT to signal it is ready
		BIT		7,A			; cmd exec bit is bit 7
		JR		NZ,Copen1	; wait until not executing
		CP      0           ; other bits zero = OK?
        JR      Z,COpenOK
        RST     2           ; Error
        DEFB    ERRCOPEN+128     ; Cannot open command channel
COpenOK:
        LD      (IX+7),CMDMAX      ; Line length (not used)
        LD      (IX+6),0        ; Position
        AND     A               ; Clear carry
		RET

CPrint:
		LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,CPrint   	; Wait for unit to become idle (not executing)
		LD		A,B				; more than 255 chars?
		CP		0
		JR		NZ,CPrintE		; then error
		LD		A,C
		CP		CMDMAX+1
		JR		NC,CprintE ; too many characters?
		LD		A,C	       ; if we got here, we have up to 253 chars in the buffer
		LD		(SDBUF),A	; char count first in buffer
		LD		DE,SDBUF+1	; destination for string in HL
		LDIR				; copy it in!
		LD		A,CMDCPRIN  ; print command
		LD		(SDCMD),A
		LD		A,128
		LD		(SDSTAT),A ; do it!
CPrint1:
		LD 		A,(SDSTAT)
		BIT 	7,A
		JR		NZ,CPrint1  ; ready?
		CP      0           ; other bits zero = OK?
        JR      NZ,CPrintE
		RET

CPrintE:
		RST		2
		DEFB	ERRCPRIN+128  ; print to cmd error

CInput:
		LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,CInput  	; Wait for unit to become idle (not executing)
		LD		A,CMDCINPU  ; input command
		LD		(SDCMD),A
		LD		A,128
		LD		(SDSTAT),A ; do it!
CInput1:
		LD 		A,(SDSTAT)
		BIT 	7,A
		JR		NZ,CInput1  ; ready?
		CP      0           ; other bits zero = OK?
        JR      NZ,CInputE ; Error! (or EOF)
Cinput1b:
		LD		A,B
		CP		0
		JR		NZ,CInputOK ; more than 255 chars in input string - very unlikely but always OK since input string never longer than 252 chars!
		LD		A,(SDBUF) ; get input string length from buffer
		CP		C		; compare to what we can fit input string in destination (one byte less, actually!)
		JP		NC,CInputE2 ; NC set if (SDBUF) >= C
CInputOK:
		EX		DE,HL
		LD		HL,SDBUF+1
		LD		A,(SDBUF)
		LD		C,A
		LD		B,0
		LDIR    ; copy over the data
		LD		A,13
		LD		(DE),A  ; end string with a CR char (VERY important so that's why it is here in the driver)
		LD		A,(SDSTAT) ; get SDSTAT again
		CP		ERREOF ;  check EOF again
		JR		Z,CInputEOF ; got to EOF ending
		AND		A	; otherwise just clear carry and return
		RET
CInputEOF:
		XOR		A  ; null out A
		SCF		; and set carry!
		RET

CInputE:
		CP		ERREOF  ; was it just EOF?
		JR		Z,Cinput1b ; then jump back to input routine
		RST		2
		DEFB	ERRCINPU+128  ; print to cmd error
CInputE2:
		RST		2
		DEFB	ERRCINLEN+128 ; input length error

CClose:
		LD		A,(SDSTAT)
		BIT		7,A
		JR		NZ,CClose  	; Wait for unit to become idle (not executing)
		LD		A,CMDCCLOS  ; input command
		LD		(SDCMD),A
		LD		A,128
		LD		(SDSTAT),A ; do it!
CClose1:
		LD 		A,(SDSTAT)
		BIT 	7,A
		JR		NZ,CClose1  ; ready?
		CP		0
		JP		Z,CClosOK
		RST		2
		DEFB	ERRCCLOS+128
CClosOK:
		AND 	A   ; clear carry
		RET

CBlockin:
CBlockout:
		RST2
		DEFB ERRNOBLK+128
		RET


printstat:
		LD		HL,SDBUF
		LD		BC,256   ; print it all
		CALL	PROUT
		RET
PrintQ0buf:
		LD		HL,Q0BUF
		LD		BC,253   ; print it all
		CALL	PROUT
		RET
TestFindQ0:
		CALL	FINDQ0
		LD		A,E
		CP		0
		JP		Z, TestFindQ0NO ; no Q0!
		PUSH	IX
		POP		HL  ; Yes we found Q0. Put in HD
		RET
TestFindQ0NO:
		LD		HL,0  ; return 0
		RET

SerAddChar:
		LD		B,0
		LD		A,(SEROUTP2)  ; load offset pointer to C
		LD		C,A
		LD		HL,SEROUTBUF  ; buffer base to HL
		ADD		HL,BC       ; find point in buffer
		LD		(HL),E    ; E comes from DE in CALL (xxx,DE) routine - char to print
		INC		C      ; increase buffer pointer, wrap to 0 if > 255
		LD		A,C
		LD		(SEROUTP2),A; store pointer
;		LD		A,(SEROUTP1) ; get send pointer P1
;		CP		C			; compare to output pointer P2
;		JP		Z,SerAddCharFull ; buffer full! jump to adjust P2 back again
		RET
;SerAddCharFull:
;		DEC		C
;		LD		A,C
;		LD		(SEROUTP2),A
;		RET

SerPrintChars:
		LD		B,0
		LD		A,(SERINP1)   ; P1 offset
		LD		C,A			 ; store to C
SerPrintCharsLoop:
		LD 		HL,SERINBUF    ; address to input buffer
		LD		A,(SERINP2)  ; P2 OFFSET - must be read for each loop turn (can increase during print!)
		CP		C				; compare
		JP		Z,SerPrintCharsEnd ; stop printing if we reached the end (two pointers the same)
		ADD		HL,BC			; form address to print
		PUSH	BC			; save BC
		LD		BC,1     ; one char to print
		CALL	PROUT		; print it! (not the most efficient way but.....)
		POP		BC			; take back counter
		INC		C			; go to next char, wrap to zero if C>255
		JP		SerPrintCharsLoop  ; loop again to print more characters
SerPrintCharsEnd:
		LD		A,C
		LD		(SERINP1),A		; store back P1 from C
		RET

SerPrintOutBuf:
		LD		HL,SEROUTBUF
		LD		BC,256
		CALL	PROUT
		RET


SerGetChar:  ; returns char in L or H>0 if no character avail
		LD		B,0
		LD		A,(SERINP1)   ; P1 offset
		LD		C,A			 ; store to C
		LD 		HL,SERINBUF    ; address to input buffer
		LD		A,(SERINP2)  ; P2 OFFSET
		CP		C				; compare
		JP		Z,SerGetCharNoChar ; stop printing if we reached the end (two pointers the same)
		ADD		HL,BC			; form address in buf
		INC		C  ; next char (wrap if end of buf)
		LD		A,C
		LD 		(SERINP1),A
		LD		A,(HL)
		LD 		H,0
		LD		L,A
		RET
SerGetCharNoChar:
		LD		H,1
		RET

USBPrintChars:
		LD		B,0
		LD		A,(USBINP1)   ; P1 offset
		LD		C,A			 ; store to C
USBPrintCharsLoop:
		LD 		HL,USBINBUF    ; address to input buffer
		LD		A,(USBINP2)  ; P2 OFFSET - must be read for each loop turn (can increase during print!)
		CP		C				; compare
		JP		Z,USBPrintCharsEnd ; stop printing if we reached the end (two pointers the same)
		ADD		HL,BC			; form address to print
		PUSH	BC			; save BC
		LD		BC,1     ; one char to print
		CALL	PROUT		; print it! (not the most efficient way but.....)
		POP		BC			; take back counter
		INC		C			; go to next char, wrap to zero if C>255
		JP		USBPrintCharsLoop  ; loop again to print more characters
USBPrintCharsEnd:
		LD		A,C
		LD		(USBINP1),A		; store back P1 from C
		RET

USBPrintOutBuf:
		LD		HL,USBOUTBUF
		LD		BC,256
		CALL	PROUT
		RET


USBAddChar:
		LD		B,0
		LD		A,(USBOUTP2)  ; load offset pointer to C
		LD		C,A
		LD		HL,USBOUTBUF  ; buffer base to HL
		ADD		HL,BC       ; find point in buffer
		LD		(HL),E    ; E comes from DE in CALL (xxx,DE) routine - char to print
		INC		C      ; increase buffer pointer, wrap to 0 if > 255
		LD		A,C
		LD		(USBOUTP2),A; store pointer
;		LD		A,(SEROUTP1) ; get send pointer P1
;		CP		C			; compare to output pointer P2
;		JP		Z,SerAddCharFull ; buffer full! jump to adjust P2 back again
		RET

;;;;;;;;;;;;;;;;;;;; testing
RANDBLIN:
		LD				(BLK),DE ; store block in DE
		LD				A,253
		LD				(BLKSZ),A    ; BASIC block size
		CALL 			FINDQ0
		LD				A,E
		CP				0
		JR				Z,RANDBLINnewq0
RANDBLINq0exist:
		LD				(IX+4),253  ; 253 bytes dim length
		LD				(IX+5),0
		LD				HL,Q0BUF
		LD				(IX+6),L   ; address to q0$ now points to Q0BUF!!
		LD				(IX+7),H
		LD				(IX+8),253  ; 253 bytes actual length
		LD				(IX+9),0
		JP				RANDBLINreadblk
RANDBLINnewq0:
		; load varlist into next pointer in "our" Q0
		LD				DE,(Varlst)
		LD 				(Q0datanext),DE
		; load varlist with pointer to our Q0
		LD				DE,Q0BUF
		LD				(Varlst),DE
		LD				DE,Q0Data
		PUSH			DE
		POP				IX ; store the pointer to the newly created Q0$ in IX
RANDBLINreadblk:         ; when we come here IX will always have a pointer to q0$
		CALL			BlockInRaw ; read in the block as set up by FNO, BLK, BLKSZ=253
		LD				HL,SDBUF ;
		LD				DE,Q0BUF
		LD				BC,253
		LDIR			; copy it over to Q0$
		RET
;;;;;;;;;;;;;;;;;;;; testing
RANDBLOUT:
		LD		(BLK),DE ; file number already in RANDFNO, data is in Q0
		CALL 	FINDQ0
		LD		A,E
		CP		0
		JR		NZ,RANDBLOUTcopyq0 ; if no Q0$, report error and quit!
		RST		2
		DEFB	ERRBLOUT+128
RANDBLOUTcopyq0:
		LD		DE,SDBUF	; destination in DE, Q0 source in IX
		LD		L,(IX+6)	; address to Q0 data in HL
		LD		H,(IX+7)	; address to Q0 data in HL
		LD		BC,253 ; 253 bytes (if Q0 is shorter we will get some bogus data)
		LDIR		; copy in the data
RANDBLOUTwrblk:     ; once Q0$ copied over to SDBUF, perform write
		LD				A,253
		LD				(BLKSZ),A    ; BASIC block size
		CALL			BlockoutRaw  ; do write
		RET

FINDQ0:					; return E=1 if Q0 found, addr to Q0$ in IX if found
		LD				IX,	Varlst
		LD				E,0  ; found flag
		LD				A,(IX)
		LD				B,(IX+1)
		OR				B
		JP				Z,FINDQ0end ; if empty varlist, just jump to end
		LD				L,(IX)	  ; if not empty, transfer pointer to IX
		LD				H,(IX+1)
		PUSH			HL
		POP				IX
FINDQ0a:							; loop through variables to see if Q0$ is there
		LD				A,(IX)
		CP				2		;  string is type 2, '0' is coded as 0000 in upper 4 bits
		JR				NZ,    FINDQ0next ; not right type or number, find next entry
		LD				A,(IX+1)
		CP				'Q'
		JR				NZ,		FINDQ0next ; not 'Q', find next
		LD				E,1		; OK we found Q0! set E flag to 1 (Q0$ ptr in IX)
		PUSH			IX
		POP				HL  ; and in HL
		JP				FINDQ0end
FINDQ0next:
		LD				L,(IX+2)
		LD				H,(IX+3)
		PUSH			HL
		POP				IX  ; new address now in IX
		LD				A,H
		OR				L
		JR				Z,FINDQ0end ; next pointer was zer0 - end of list!
		JP				FINDQ0a ; jump again
FINDQ0end:
		RET
RANDSETFNO:			; set the file number
		LD				(SDFNO),DE
		RET
SETEXEC:
		LD			(EXECADDR),DE   ; set exec address for loadfile
		RET

LOADFILE:	;	file should be open wiht BASIC "open" and SDNFO set and EXECADDR set, DE has target address
		LD				A,253
		LD				(BLKSZ),A ;  and in blk as well
		LD				A,0
		LD				(BLK),A  ; start block is zero
LOADFILEb:
		LD				HL,SDBUF  ; get data from SDBUF
		LD				BC,253  ; default block size
		CALL			Blockinraw  ; will exit to basic error routine if error so no need to check anything
		LDIR			; transfer the block
		LD				A,(SDSTAT)  ; check the status of the command
		CP				0  ; all OK? (EOF is also an error)
		JP				Z,LOADFILEb
		CP				ERREOF ; EOF means we got the file and should (maybe) execute it
		JP				Z,LOADFILEexec
		RST				2
		DEFB			ERRLOAD+128
LOADFILEexec:
		LD				HL,(EXECADDR)
		LD				A,H
		OR				L
		CP				0  ; exec addr 0?
		RET				Z  ; yes, return
		JP				(HL) ; no, jump to it

; orgs commented away to test UFD-dos functionality (t
;		org				RANDJPTAB-64  ; to safely be able to JR to these routines
RANDJPSETFNO:
		JP				RANDSETFNO
RANDJPBLIN:
		JP				RANDBLIN ; jump to real routine
RANDJPBLOUT:
		JP				RANDBLOUT ; jump to real routine

;		org 			RANDJPTAB
		JR				RANDJPSETFNO
		JR				RANDJPBLIN   ; jump to jump routine!
		JR				RANDJPBLOUT  ; jump to jump routine!

		org 24573
UFDDOSINIT:
		RET
		RET
		RET				; three ret instructions, can be patched
		;				; with jump to ufddos initroutine if needed
       ; JP	UFDINITADDR  ;  // in ufd-dos this is c3 43 65
		END
